import logging
import requests
import json
import base64
from flask import redirect, Blueprint, url_for, request, session
from flask_login import login_user
from . import User
from .. import config, utils
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
logger = logging.getLogger('sigmund')
google_login_blueprint = Blueprint('google_login', __name__)


def get_google_provider_cfg():
    return requests.get(config.google_discovery_url).json()
    
    
@google_login_blueprint.route("/")
def login():
    if not config.google_login_enabled:
        return jsonify(success=False, message="Login method disabled"), 403
    from oauthlib.oauth2 import WebApplicationClient
    client = WebApplicationClient(config.google_client_id)
    google_provider_cfg = get_google_provider_cfg()
    authorization_endpoint = google_provider_cfg["authorization_endpoint"]
    request_uri = client.prepare_request_uri(
        authorization_endpoint,
        redirect_uri=request.base_url.replace('http:', 'https:') + "callback",
        scope=["openid", "email", "profile"],
    )
    return redirect(request_uri)


@google_login_blueprint.route("/callback")
def callback():
    from oauthlib.oauth2 import WebApplicationClient
    client = WebApplicationClient(config.google_client_id)
    # Get authorization code Google sent back to you
    code = request.args.get("code")
    # Find out what URL to hit to get tokens that allow you to ask for
    # things on behalf of a user
    google_provider_cfg = get_google_provider_cfg()
    token_endpoint = google_provider_cfg["token_endpoint"]
    # Prepare and send a request to get tokens! Yay tokens!
    try:
        token_url, headers, body = client.prepare_token_request(
            token_endpoint,
            authorization_response=request.url.replace('http:', 'https:'),
            redirect_url=request.base_url.replace('http:', 'https:'),
            code=code
        )
    except Exception as e:
        logger.warning(f'failed to prepare token request: {e}')
        return redirect('/')
        
    token_response = requests.post(
        token_url,
        headers=headers,
        data=body,
        auth=(config.google_client_id, config.google_client_secret),
    )
    # Parse the tokens!
    client.parse_request_body_response(json.dumps(token_response.json()))    
    # Now that you have tokens (yay) let's find and hit the URL
    # from Google that gives you the user's profile information,
    # including their Google profile image and email
    userinfo_endpoint = google_provider_cfg["userinfo_endpoint"]
    uri, headers, body = client.add_token(userinfo_endpoint)
    userinfo_response = requests.get(uri, headers=headers, data=body)
    # You want to make sure their email is verified.
    # The user authenticated with Google, authorized your
    # app, and now you've verified their email through Google!
    if not userinfo_response.json().get("email_verified"):
        logger.info(f'google log-in failed ({username})')
        return redirect('/login_failed')
    unique_id = userinfo_response.json()["sub"]
    username = userinfo_response.json()["given_name"]
    logger.info(f'google log-in successful ({username})')
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                     length=32,
                     salt=config.encryption_salt,
                     iterations=100000,
                     backend=default_backend())
    session['encryption_key'] = base64.urlsafe_b64encode(
        kdf.derive(unique_id.encode()))
    user = User(f'{username} (google)')
    login_user(user)
    logger.info(f'initializing encryption key')    
    return redirect('/')



import logging
import os
logger = logging.getLogger('sigmund')
# Depending on whether we are running inside a Flask app or not, we need to
# instantiate the database model differently. It is difficult to determine
# automatically whether Flask is running, because at this point we're still
# in the initialization phase. Therefore, we're explicitly setting an 
# environment variable when creating the Flask app.
if os.environ.get('USE_FLASK_SQLALCHEMY', False):
    logger.info('using flask_sqlachemy')
    from flask_sqlalchemy import SQLAlchemy as _BaseSQLAlchemy
    db = _BaseSQLAlchemy()
    Column = db.Column
    Integer = db.Integer
    String = db.String
    ForeignKey = db.ForeignKey
    LargeBinary = db.LargeBinary
    DateTime = db.DateTime
    Model = db.Model
else:
    from sqlalchemy import create_engine, Column, Integer, String, \
        ForeignKey, LargeBinary, DateTime
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import scoped_session, sessionmaker
    import logging
    logger.info('using standalone_sqlachemy')
    engine = create_engine('sqlite:///:memory:')
    session = scoped_session(sessionmaker(autocommit=False, autoflush=False,
                                          bind=engine))
    Base = declarative_base()
    # Create a db object that mimics Flask-SQLAlchemy's, with Model as an 
    # attribute
    class SQLAlchemy:
        def __init__(self, metadata):
            self.Model = declarative_base(metadata=metadata)
    db = SQLAlchemy(metadata=Base.metadata)
    db.session = session
    Model = db.Model
    Model.query = session.query_property()
    def init_db(): Model.metadata.create_all(bind=engine)
    def drop_db(): Model.metadata.drop_all(bind=engine)


class User(Model):
    __tablename__ = 'user'

    user_id = Column(Integer, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    active_conversation_id = Column(
        Integer, ForeignKey('conversation.conversation_id'))


class Conversation(Model):
    __tablename__ = 'conversation'

    conversation_id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.user_id'))
    # Data corresponds to an encrypted blob. After encryption, it becomes a
    # JSON string with the following format:
    # {
    #     "condensed_text": "A summary",
    #     "message_history": [
    #         message_id1,
    #         message_id2
    #     ],
    #     "condensed_message_history": [
    #         { message object1 },
    #         { message object2 }
    #     ]
    #     'title': self._conversation_title,
    # }
    # The message_ids refer to entries in the message table below. For older
    # conversations, the message_ids may instead by dicts. The 
    # DatabaseManager.get_message() function takes care of this.
    data = Column(LargeBinary)


class Message(Model):
    __tablename__ = 'message'

    message_id = Column(Integer, primary_key=True)
    conversation_id = Column(Integer,
                             ForeignKey('conversation.conversation_id'))
    # Data corresponds to an encrypted blob. After encryption, it becomes a
    # JSON string.
    data = Column(LargeBinary)


class Attachment(Model):
    __tablename__ = 'attachment'

    attachment_id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.user_id'))
    data = Column(LargeBinary)


class Activity(Model):
    __tablename__ = 'activity'

    activity_id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.user_id'))
    time = Column(DateTime)
    tokens_consumed = Column(Integer)


class Subscription(Model):
    __tablename__ = 'subscription'

    subscription_id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.user_id'))
    from_date = Column(DateTime)
    to_date = Column(DateTime)
    stripe_subscription_id = Column(String(80))
    stripe_customer_id = Column(String(80))


class Setting(Model):
    __tablename__ = 'setting'

    setting_id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.user_id'))
    key = Column(String(80))
    value = Column(String(80))



import logging
logger = logging.getLogger('sigmund')


db_initialized = False


def predict(prompt, model: str, json: bool = False,
            max_json_retry: int = 5) -> str:
    """Provides a static interface to predict a single response."""
    from .database.models import init_db
    from .model import model as sigmund_model
    global db_initialized
    if not db_initialized:
        init_db()
        db_initialized = True
    model = sigmund_model(None, model)
    if not json:
        return model.predict(prompt)
    import json
    model.json_mode = True
    for _ in range(max_json_retry):
        try:
            return json.loads(model.predict(prompt))
        except Exception as e:
            logger.warning(f'failed to predict JSON: {e}')
            continue
    raise ValueError(f'Failed after {max_json_retry} attempts')



import logging
from flask import jsonify, Blueprint, request, Response
from flask_cors import cross_origin
from .. import config, utils, prompt
from ..sigmund import Sigmund
from .app import get_sigmund
logger = logging.getLogger('sigmund')
public_blueprint = Blueprint('public', __name__)


@public_blueprint.route('/search', methods=['POST'])
@cross_origin()
def search():
    data = request.json
    config.db_cache = data.get('source', 'default')
    query = data.get('query', None)
    offset = data.get('offset', 0) * config.public_search_docs_max
    sigmund = Sigmund(user_id='dummy', persistent=False, encryption_key=None)
    sigmund.documentation.clear()
    sigmund.documentation.search([query])
    docs = []
    urls = []
    logging.info(f'public search: {query} (offset={offset})')
    for doc in sigmund.documentation._documents[offset:]:
        if doc.metadata['url'] in urls:
            print('skipping duplicate')
            continue
        urls.append(doc.metadata['url'])
        doc.page_content = doc.page_content[
            :config.public_search_max_doc_length]
        docs.append(doc)
        if len(docs) >= config.public_search_docs_max:
            break
    sigmund.documentation._documents = docs
    logging.info(f'public search results: {len(docs)}')
    query = prompt.render(prompt.PUBLIC_SEARCH_PROMPT,
                          documentation=sigmund.documentation)
    results = utils.md(
        sigmund.public_model.predict(query).replace('\n\n-', '\n-'))
    return jsonify(results=results)


@public_blueprint.route('/search_widget')
def search_widget():
    return utils.render('search-widget.html')


@public_blueprint.route('/search-widget.js')
@cross_origin()
def search_widget_js():
    return Response(utils.render('search-widget.js'),
                    mimetype='text/javascript')



import logging
import os
import io
import subprocess
import tempfile
from pdfminer.high_level import extract_text
from . import config, prompt
logger = logging.getLogger('sigmund')


def file_to_text(name: str, content: bytes, model) -> str:
    """Generates a text representation of a file, based on its name and 
    content. If the file cannot be converted to a text representation, 
    'No description' is returned. The representation is limited to the maximum
    length as specified in the config.
    """
    suffix = os.path.splitext(name)[1].lower()
    if suffix in ['.txt', '.md']:
        # File formats that don't need any processing
        text_representation = content.decode('utf-8', errors='ignore')
    elif suffix == '.pdf':
        try:
            text_representation = extract_text(io.BytesIO(content))
        except Exception as e:
            logger.error(f'failed to extract text: {e}')
            return 'No description'
    else:
        with tempfile.NamedTemporaryFile(suffix=suffix) as temp_file:
            temp_file.write(content)
            try:
                output = subprocess.run(
                    ['pandoc', '-s', temp_file.name, '-t', 'plain'],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    check=True,
                    text=True
                )
            except subprocess.CalledProcessError as e:
                logger.error(f'failed to extract text: {e}')
                return 'No description'
            except FileNotFoundError as e:
                logger.error(f'pandoc is not available: {e}')
                return 'No description'
        text_representation = output.stdout.strip()
    if not text_representation:
        text_representation = content.decode('utf-8', errors='ignore')
    text_representation = text_representation.strip()[
        :config.max_text_representation_length]
    if len(text_representation) > config.text_respresentation_summary_threshold:
        original_length = len(text_representation)
        text_representation = model.predict(
            prompt.render(prompt.SUMMARIZE_PROMPT,
                          text_representation=text_representation))
        logger.info(f'summarized from {original_length} to {len(text_representation)}')
    if not text_representation:
        return 'No description'
    return text_representation


def describe_file(name: str, content: bytes, model) -> str:
    """Generates an LLM description of a file based on its name and content."""
    return model.predict(
        prompt.render(prompt.DESCRIBE_PROMPT, name=name,
                      text_representation=content))


def attachments_prompt(db) -> str:
    """Returns a prompt fragment to include in the system prompt, containing
    a list of all attachments including descriptions. An empty string
    is returned if there are no attachments.
    """
    description = []
    for attachment in db.list_attachments().values():
        description.append(
            f'- {attachment["filename"]}: {attachment["description"]}')
    if not description:
        return ''
    return prompt.render(prompt.SYSTEM_PROMPT_ATTACHMENTS,
                         description='\n'.join(description))



import logging
import json
import zlib
import uuid
import time
import re
from pathlib import Path
from cryptography.fernet import InvalidToken
from .model import model
from . import prompt, config, utils, attachments
from langchain.schema import HumanMessage, AIMessage, SystemMessage, \
    FunctionMessage
logger = logging.getLogger('sigmund')


class Messages:
    
    def __init__(self, sigmund, persistent=False):
        self._sigmund = sigmund
        self._persistent = persistent
        if self._persistent:
            self.load()
        else:
            self.init_conversation()
        
    def __len__(self):
        return len(self._message_history)
        
    def __iter__(self):
        for msg in self._message_history:
            yield msg
            
    def init_conversation(self):
        self._condensed_text = None
        metadata = self.metadata()
        metadata['answer_model'] = 'welcome-bot'
        self._conversation_title = config.default_conversation_title
        self._message_history = [['assistant', self.welcome_message(), 
                                  metadata]]
        self._condensed_message_history = [
            [role, content] for role, content, metadata
            in self._message_history[:]]        
            
    def metadata(self, message_id=None):
        return {'message_id': str(uuid.uuid4()) 
                    if message_id is None else message_id,
                'timestamp': utils.current_datetime(),
                'sources': self._sigmund.documentation.to_json(),
                'search_model': self._sigmund.model_config['search_model'],
                'condense_model': self._sigmund.model_config['condense_model'],
                'answer_model': self._sigmund.model_config['answer_model']}
        
    def append(self, role, message, message_id=None):
        metadata = self.metadata(message_id=message_id)
        self._message_history.append([role, message, metadata])
        self._condensed_message_history.append([role, message])
        self._condense_message_history()
        if self._persistent:
            self.save()
        return metadata
    
    def delete(self, message_id):
        message_to_remove = None
        condensed_message_to_remove = None
        for role, message, metadata in self._message_history:
            if metadata['message_id'] == message_id:
                message_to_remove = [role, message, metadata]
                condensed_message_to_remove = [role, message]
                break
        else:
            logger.info(f'message not found for deletion: {message_id}')
            return
        if message_to_remove:
            logger.info(f'deleting message: {message_id}')
            self._message_history.remove(message_to_remove)
        if condensed_message_to_remove:
            try:
                self._condensed_message_history.remove(
                    condensed_message_to_remove)
            except ValueError:
                logger.error(
                    f'Could not find condensed message to remove for {message_id}')
        if self._persistent:
            self.save()
            
    def prompt(self, system_prompt=None, skip_large_tool_results=False):
        """The prompt consists of the system prompt followed by a sequence of
        AI, user, and tool/ function messages.
        
        If no system prompt is provided, one is automatically constructed.
        Typically, an explicit system_prompt is provided during the search
        phase, but not during the answer phase.
        """
        if system_prompt is None:
            system_prompt = self._system_prompt()
        model_prompt = [SystemMessage(content=system_prompt)]
        msg_len = len(self._condensed_message_history)
        for msg_nr, (role, content) in enumerate(
                self._condensed_message_history):
            if role == 'assistant':
                model_prompt.append(AIMessage(content=content))
            elif role == 'user':
                model_prompt.append(HumanMessage(content=content))
            elif role == 'tool':
                if msg_nr + config.keep_tool_results < msg_len:
                    continue
                if len(content) > config.large_tool_result_length:
                    continue
                model_prompt.append(FunctionMessage(content=content,
                                                    name='tool_function'))
            else:
                raise ValueError(f'Invalid role: {role}')
        return model_prompt
    
    def visible_messages(self):
        """Yields role, message, metadata while ignoring messages and 
        converting tool messages into user messages with tool result as 
        content. This is mainly for display in the web interface.
        """
        for role, message, metadata in self:
            if role == 'tool':
                role = 'assistant'
                tool_results = json.loads(message)
                message = tool_results['content']
            if not message.strip():
                continue
            yield role, message, metadata

    def welcome_message(self):
        return config.welcome_message
        
    def _condense_message_history(self):
        system_prompt = self._system_prompt()
        messages = [{"role": "system", "content": system_prompt}]
        prompt_length = sum(len(content) for role, content
                            in self._condensed_message_history
                            if role != 'tool')
        logger.info(f'system prompt length: {len(system_prompt)}')
        logger.info(f'prompt length (without system prompt): {prompt_length}')
        if prompt_length <= config.max_prompt_length:
            logger.info('no need to condense')
            return
        condense_length = 0
        condense_messages = []
        while condense_length < config.condense_chunk_length:
            role, content = self._condensed_message_history.pop(0)
            if role == 'tool' and \
                    len(content) > config.large_tool_result_length:
                continue
            condense_length += len(content)
            condense_messages.insert(0, (role, content))
        logger.info(f'condensing {len(condense_messages)} messages')
        condense_prompt = prompt.render(
            prompt.CONDENSE_HISTORY,
            history=''.join(
                f'You said: {content}' if role == 'system' 
                else f'User said: {content}'
                for role, content in condense_messages))
        self._condensed_text = self._sigmund.condense_model.predict(
            condense_prompt)
        
    def _system_prompt(self):
        """The system prompt that is used for question answering consists of
        several fragments.
        """
        # There is always and identity and a list of attached files
        if self._sigmund.search_first:
            system_prompt = [prompt.SYSTEM_PROMPT_IDENTITY_WITH_SEARCH]
        else:
            system_prompt = [prompt.SYSTEM_PROMPT_IDENTITY_WITHOUT_SEARCH]
        system_prompt.append(
            attachments.attachments_prompt(self._sigmund.database))
        # For models that support this, there is also an instruction indicating
        # that a special marker can be sent to indicate that the response isn't
        # done yet. Not all models support this to avoid infinite loops.
        if self._sigmund.answer_model.supports_not_done_yet:
            system_prompt.append(prompt.SYSTEM_PROMPT_NOT_DONE_YET)
        # If available, documentation is also included in the prompt
        if len(self._sigmund.documentation):
            system_prompt.append(self._sigmund.documentation.prompt())
        # And finally, if the message history has been condensed, this is also
        # included.
        if self._condensed_text:
            logger.info('appending condensed text to system prompt')
            system_prompt.append(prompt.render(
                prompt.SYSTEM_PROMPT_CONDENSED,
                summary=self._condensed_text))
        # Combine all non-empty prompt chunks
        return '\n\n'.join(chunk for chunk in system_prompt if chunk.strip())
        
    def _update_title(self):
        """The conversation title is updated when there are at least two 
        messages, excluding the system prompt and AI welcome message. Based on
        the last messages, a summary title is then created.
        """
        if len(self) <= 2 or \
                self._conversation_title != config.default_conversation_title:
            return
        title_prompt = [SystemMessage(content=prompt.TITLE_PROMPT)]
        title_prompt += self.prompt()[2:]
        self._conversation_title = self._sigmund.condense_model.predict(
            title_prompt).strip('"\'')
        if len(self._conversation_title) > 100:
            self._conversation_title = self._conversation_title[:100] + '…'

    def load(self):
        try:
            conversation = self._sigmund.database.get_active_conversation()
        except InvalidToken as e:
            self.init_conversation()
            self.save()
            logger.error(f'failed to get active conversation: {e}')
            return
        if not conversation['message_history']:
            self.init_conversation()
            return
        modified = False
        for _, _, metadata in conversation['message_history']:
            if 'message_id' not in metadata:
                metadata['message_id'] = str(uuid.uuid4())
                modified = True
        self._conversation_title = conversation['title']
        self._message_history = conversation['message_history']
        self._condensed_text = conversation['condensed_text']
        self._condensed_message_history = \
            conversation['condensed_message_history']
        if self._persistent and modified:
            self.save()
    
    def save(self):
        self._update_title()
        conversation = {
            'condensed_text': self._condensed_text,
            'message_history': self._message_history,
            'condensed_message_history': self._condensed_message_history,
            'title': self._conversation_title,
        }
        self._sigmund.database.update_active_conversation(conversation)



from ._base_tool import BaseTool
from ._search_documentation import search_documentation
from ._search_google_scholar import search_google_scholar
from ._read_attachment import read_attachment
from ._execute_code import execute_code
from ._download import download
from ._generate_image import generate_image



from ._base_model import BaseModel


def model(sigmund, model, **kwargs):
    """A factory function that returns a Model instance."""
    if model == 'o1':
        from ._openai_model import OpenAIModel
        return OpenAIModel(sigmund, 'o1', **kwargs)
    if model in ('gpt-4', 'gpt-4o'):
        from ._openai_model import OpenAIModel
        return OpenAIModel(sigmund, 'gpt-4o', **kwargs)
    if model == 'gpt-4o-mini':
        from ._openai_model import OpenAIModel
        return OpenAIModel(sigmund, 'gpt-4o-mini', **kwargs)
    if model == 'gpt-3.5':
        from ._openai_model import OpenAIModel
        return OpenAIModel(sigmund, 'gpt-3.5-turbo', **kwargs)
    if model == 'claude-2.1':
        from ._anthropic_model import AnthropicModel
        return AnthropicModel(sigmund, 'claude-2.1', **kwargs)
    if model == 'claude-3-opus':
        from ._anthropic_model import AnthropicModel
        return AnthropicModel(sigmund, 'claude-3-opus-20240229', **kwargs)
    if model == 'claude-3-sonnet':
        from ._anthropic_model import AnthropicModel
        return AnthropicModel(sigmund, 'claude-3-sonnet-20240229', **kwargs)
    if model == 'claude-3.5-sonnet':
        from ._anthropic_model import AnthropicModel
        return AnthropicModel(sigmund, 'claude-3-5-sonnet-latest', **kwargs)
    if model == 'claude-3-haiku':
        from ._anthropic_model import AnthropicModel
        return AnthropicModel(sigmund, 'claude-3-haiku-20240307', **kwargs)
    if model == 'claude-3.5-haiku':
        from ._anthropic_model import AnthropicModel
        return AnthropicModel(sigmund, 'claude-3-5-haiku-latest', **kwargs)
    if 'mistral' in model:
        from ._mistral_model import MistralModel
        return MistralModel(sigmund, model, **kwargs)
    if model == 'dummy':
        from ._dummy_model import DummyModel
        return DummyModel(sigmund, **kwargs)
    raise ValueError(f'Unknown model: {model}')



import json
import logging
from werkzeug.utils import secure_filename
from flask import request, jsonify, Response, redirect, session, \
    stream_with_context, make_response, Blueprint
from flask_login import login_required
from .. import config, utils, attachments
from ..sigmund import Sigmund
from .app import get_sigmund
from redis import Redis
logger = logging.getLogger('sigmund')
api_blueprint = Blueprint('api', __name__)
redis_client = Redis()


@api_blueprint.route('/chat/start', methods=['POST'])
@login_required
def api_chat_start():
    data = request.json
    session['user_message'] = data.get('message', '')
    session['message_id'] = data.get('message_id', None)
    sigmund = get_sigmund()
    redis_client.delete(f'stream_cancel_{sigmund.user_id}')
    return '{}'
    

@api_blueprint.route('/chat/cancel', methods=['POST'])
@login_required
def api_chat_cancel_stream():
    sigmund = get_sigmund()
    logger.info(f'cancelling stream for {sigmund.user_id}')
    redis_client.set(f'stream_cancel_{sigmund.user_id}', '1')
    return jsonify({'status': 'cancelled'}), 200


@api_blueprint.route('/chat/stream', methods=['GET'])
@login_required
def api_chat_stream():
    message = session['user_message']
    message_id = session['message_id']
    sigmund = get_sigmund()
    logger.info(f'starting stream for {sigmund.user_id}')
    def generate():
        for reply, metadata in sigmund.send_user_message(message, message_id):
            if isinstance(reply, dict):
                reply = json.dumps(reply)
            else:
                reply = json.dumps(
                    {'response': utils.md(
                        f'{config.ai_name}: {config.process_ai_message(reply)}'),
                     'metadata': metadata})
            logger.debug(f'ai message: {reply}')
            yield f'data: {reply}\n\n'
            if redis_client.get(f'stream_cancel_{sigmund.user_id}'):
                logger.info(f'stream cancelled for {sigmund.user_id}')
                break
        yield 'data: {"action": "close"}\n\n'
    return Response(stream_with_context(generate()),
                    mimetype='text/event-stream')
    
    
@api_blueprint.route('/conversation/new')
@login_required
def new_conversation():
    sigmund = get_sigmund()
    sigmund.database.new_conversation()
    return redirect('/chat')


@api_blueprint.route('/conversation/clear')
@login_required
def clear_conversation():
    sigmund = get_sigmund()
    sigmund.messages.init_conversation()
    sigmund.messages.save()
    return redirect('/chat')


@api_blueprint.route('/conversation/activate/<int:conversation_id>',
                     methods=['GET', 'POST'])
@login_required
def activate_conversation(conversation_id):
    if conversation_id:
        sigmund = get_sigmund()
        sigmund.database.set_active_conversation(conversation_id)
    logger.info('redirecting to chat')
    return redirect('/chat')


@api_blueprint.route('/conversation/list', methods=['GET'])
@login_required
def list_conversations():
    sigmund = get_sigmund()
    query = request.args.get('query', None)
    return jsonify(sigmund.database.list_conversations(query))
    

@api_blueprint.route('/conversation/delete/<int:conversation_id>',
                     methods=['DELETE'])
@login_required
def delete_conversation(conversation_id):
    sigmund = get_sigmund()
    sigmund.database.delete_conversation(conversation_id)
    return '', 204


@api_blueprint.route('/attachments/list', methods=['GET', 'POST'])
@login_required
def list_attachments():
    sigmund = get_sigmund()
    return jsonify(sigmund.database.list_attachments())


@api_blueprint.route('/attachments/delete/<int:attachment_id>',
                     methods=['DELETE'])
@login_required
def delete_attachment(attachment_id):
    sigmund = get_sigmund()
    success = sigmund.database.delete_attachment(attachment_id)
    return jsonify(success=success)
    

@api_blueprint.route('/attachments/add', methods=['POST'])
@login_required
def add_attachment():
    sigmund = get_sigmund()
    if 'file' not in request.files:
        return jsonify(success=False, message="No file part"), 400
    file = request.files['file']
    if file.filename == '':
        return jsonify(success=False, message="No selected file"), 400
    filename = secure_filename(file.filename)
    content = attachments.file_to_text(filename, file.read(),
                                       sigmund.condense_model)
    file.seek(0)
    description = attachments.describe_file(filename, content,
                                            sigmund.condense_model)
    attachment_data = {'filename': filename,
                       'content': content,
                       'description': description}
    attachment_id = sigmund.database.add_attachment(attachment_data)
    if attachment_id == -1:
        return jsonify(success=False, message="Failed to add attachment"), 500
    else:
        return jsonify(success=True, attachment_id=attachment_id)


@api_blueprint.route('/attachments/get/<int:attachment_id>', methods=['GET'])
@login_required
def get_attachment(attachment_id):
    sigmund = get_sigmund()
    attachment_data = sigmund.database.get_attachment(attachment_id)
    if attachment_data:
        response = make_response(attachment_data['content'])
        response.headers['Content-Type'] = 'application/octet-stream'
        response.headers['Content-Disposition'] = \
            f'attachment; filename={attachment_data["filename"]}'
        return response
    return jsonify(success=False,
                   message="Attachment not found or access denied"), 404


@api_blueprint.route('/message/delete/<message_id>', methods=['DELETE'])
@login_required
def delete_message(message_id):
    sigmund = get_sigmund()
    sigmund.messages.delete(message_id)
    return jsonify(success=True)


@api_blueprint.route('/setting/set', methods=['POST'])
@login_required
def set_setting():
    data = request.json
    if not isinstance(data, dict):
        logger.warning(f'invalid setting data: {data}')
        return jsonify(success=False, message='invalid setting data')
    sigmund = get_sigmund()
    logger.info(f'setting: {data}')
    for key, value in data.items():
        sigmund.database.set_setting(key, value)
    return jsonify(success=True)
    

@api_blueprint.route('/setting/get/<key>')
@login_required
def get_setting(key):
    sigmund = get_sigmund()
    value = sigmund.database.get_setting(key)
    if value is None:
        if key in config.settings_default:
            value = config.settings_default[key]
            logger.info(f'using default setting for {key}: {value}')
        else:
            return jsonify(success=False, value=value)
    logger.info(f'get setting: {key} = {value}')
    return jsonify(success=True, value=value)
    



import json
from pathlib import Path
from langchain_openai import OpenAIEmbeddings
from langchain_community.vectorstores import FAISS
from langchain_community.vectorstores.utils import DistanceStrategy
from langchain_core.documents import Document
import logging
from . import config
from . import prompt
logger = logging.getLogger('sigmund')


class Documentation:
    
    def __init__(self, sigmund, sources=[]):
        self._sigmund = sigmund
        self._documents = []
        self._sources = sources
        
    def _doc_to_str(self, doc):
        s = '<document>\n'
        if doc.metadata.get('title', None):
            s += f'# {doc.metadata["title"]}\n\n'
        if doc.metadata.get('url', None):
            s += f'Source: {doc.metadata["url"]}\n\n'
        return s + doc.page_content + '\n</document>'
        
    def __str__(self):
        if not self._documents:
            return ''
        return '\n\n'.join(self._doc_to_str(doc) for doc in self._documents)
        
    def to_json(self):
        return json.dumps([{'page_content': doc.page_content,
                            'url': doc.metadata.get('url', None)}
                           for doc in self])
        
    def __iter__(self):
        return (doc for doc in self._documents)
    
    def __len__(self):
        return sum(len(doc.page_content) for doc in self._documents)
        
    def __contains__(self, doc):
        return doc in self._documents
    
    def prompt(self):
        if not self._documents:
            return None
        return f'''# Documentation

You have retrieved the following documentation to answer the user's question:

<documentation>
{str(self)}
</documentation>'''
    
    def append(self, doc):
        if any(doc.page_content == d.page_content for d in self):
            return
        self._documents.append(doc)
            
    def strip_irrelevant(self, question):
        important = [doc for doc in self
                     if doc.metadata.get('important', False)]
        optional = [doc for doc in self
                    if not doc.metadata.get('important', False)]
        while optional and len(important) < config.search_docs_max:
            batch = optional[:config.search_docs_max]
            logger.info(f'evaluating document batch (n={len(batch)}): {[doc.metadata["seq_num"] for doc in batch]}')
            optional = optional[config.search_docs_max:]
            prompts = [prompt.render(prompt.JUDGE_RELEVANCE,
                                     documentation=doc.page_content,
                                     question=question)
                       for doc in batch]
            self._sigmund.condense_model.json_mode = True
            replies = self._sigmund.condense_model.predict_multiple(prompts)
            self._sigmund.condense_model.json_mode = False
            # config.mistral_kwargs = dict()
            for reply, doc in zip(replies, batch):
                doc_desc = f'{doc.metadata["url"]} ({doc.metadata["seq_num"]})'
                try:
                    reply = json.loads(reply)
                except json.JSONDecodeError:
                    logger.warning(f'invalid JSON for {doc_desc} (reply: {reply})')
                    continue
                if not isinstance(reply, dict):
                    logger.warning(f'invalid JSON for {doc_desc} (reply: {reply})')
                    continue
                if not reply.get('relevant', False):
                    logger.info(f'stripping {doc_desc} (reply: {reply})')
                    continue
                logger.info(f'keeping {doc_desc} (reply: {reply})')
                important.append(doc)
        self._documents = important
    
    def clear(self):
        logger.info('clearing documentation')
        self._documents = []
        
    def search(self, queries):
        for source in self._sources:
            for doc in source.search(queries):
                if doc not in self._documents:
                    self._documents.append(doc)
        

class BaseDocumentationSource:
    
    def __init__(self, sigmund):
        self._sigmund = sigmund
    
    def search(self, queries):
        pass

    
class FAISSDocumentationSource(BaseDocumentationSource):
    
    def __init__(self, sigmund):
        super().__init__(sigmund)
        if config.openai_api_key is None:
            logger.warning(
                'No OpenAI API key provided, no documentation available')
            return
        self._embeddings_model = OpenAIEmbeddings(
            openai_api_key=config.openai_api_key,
            model=config.search_embedding_model)
        db_cache = config.db_cache_sources[config.db_cache]
        logger.info(f'reading FAISS documentation cache from {db_cache}')
        self._db = FAISS.load_local(Path(db_cache), self._embeddings_model,
                                    allow_dangerous_deserialization=True)
        if config.search_docs_distance_metric == 'cosine':
            self._db.distance_strategy = DistanceStrategy.COSINE
        elif config.search_docs_distance_metric == 'euclidean_distance':
            self._db.distance_strategy = DistanceStrategy.EUCLIDEAN_DISTANCE
        else:
            raise ValueError(
                f'invalid search_docs_distance_metric: {config.search_docs_distance_metric}')

    def search(self, queries):
        if config.openai_api_key is None:
            return []
        docs = []
        for query in queries:
            if config.log_replies:
                logger.info(f'documentation search query: {query}')
            for doc, score in self._db.similarity_search_with_score(
                    query, k=config.search_docs_per_query):
                doc_desc = f'{doc.metadata["url"]} ({doc.metadata["seq_num"]})'
                if any(doc.page_content == ref.page_content for ref, _ in docs):
                    continue
                if doc.page_content not in self._sigmund.documentation and \
                        doc.page_content not in docs:
                    docs.append((doc, score))
        # The documents that are retrieved need a little re-ordering to be
        # optimal. First we rank them by score across all search queries. Then
        # we reorder them so that documents without URL, which tend to be short
        # how-to's alternate with documents with URL, which tend to be longer
        # pieces of documentation. This is to make sure that documentation
        # stands a chance against the how-to's, which tend to be preferred by
        # the algorithm. Finally we keep only the top few documents.
        docs = sorted(docs, key=lambda doc: doc[1])
        with_url = [d for d in docs if d[0].metadata['url'] is not None]
        without_url = [d for d in docs if d[0].metadata['url'] is None]
        reordered_docs = []
        while with_url or without_url:
            if with_url:
                reordered_docs.append(with_url.pop(0))
            if without_url:
                reordered_docs.append(without_url.pop(0))
        for doc, score in reordered_docs:
            doc_desc = f'{doc.metadata["url"]} ({doc.metadata["seq_num"]})'
            logger.info(f'considering {doc_desc}, score: {score}')
        return [doc for doc, _ in reordered_docs]
